//
// HARL 2012
// Kinect depth pixel to color pixel mapping
//

// Using Nicolas Burrus tools for calibration
// http://nicolas.burrus.name/index.php/Research/KinectRgbDemoV4?from=Research.KinectRgbDemo

//---------------------------------------------------------

//
// Calibration data
//
//
// Intrinsic parameters
//  |  fx   0  cx  |
//  |   0  fy  cy  |
//  |   0   0   1  |
//
// stored as float[9] = { fx, 0, cx, 0, fy, cy, 0, 0, 1 } 
//
// Distortion coeff float[5] = { k_1, k_2, p_1, p_2, k_3 }
// look at http://opencv.willowgarage.com/documentation/camera_calibration_and_3d_reconstruction.html#index-13
//

const float rgbIntrinsics[9] = 
{ 
	5.2161910696979987e+02, 0., 3.1755491910920682e+02, 
	0., 5.2132946256749767e+02, 2.5921654718027673e+02, 
	0., 0., 1.
};

const float rgbDistortion[5] =
{
	2.5673002693536984e-01, -9.3976085633794137e-01,     	-1.8605549188751580e-03, -2.2232238578189420e-03,
	1.2453643444153988e+00
};
		
const float depthIntrinsics[9] = 
{ 
	5.8818670481438744e+02, 0., 3.1076280589210484e+02,
	0., 5.8724220649505514e+02, 2.2887144980135292e+02, 
	0., 0., 1.
};

const float depthDistortion[5] =
{
	-1.8932947734719333e-01, 1.1358015104098631e+00,
	-4.4260345347128536e-03, -5.4869578635708153e-03,
	-2.2460143607712921e+00
};

// 3x3 rotation matrix
//  |  r11  r12  r13  |
//  |  r21  r22  r23  |
//  |  r31  r32  r33  |
//
// stored as float[] = { r11, r12, r13, r21, r22, r23, r31, r32, r33 } 
//

// inv(R)
const float RDepthAxisToRgbAxis[9] =
{
	 9.9996518012567637e-01,  2.6765126468950343e-03, -7.9041012313000904e-03,
	-2.7409311281316700e-03,  9.9996302803027592e-01, -8.1504520778013286e-03,
	 7.8819942130445332e-03,  8.1718328771890631e-03,  9.9993554558014031e-01
};

// inv(T)
const float TDepthAxisToRgbAxis[3] =
{
	-2.5558943178152542e-02, 1.0109636268061706e-04, 2.0318321729487039e-03
};

//---------------------------------------------------------

/**
	Get intrinsic parameters from 3x3 intrinsic matrix.
*/
void getIntrinsicParameters( float *fx, float *fy, float *cx, float *cy, 
	const float *intrinsics)
{
	*fx = intrinsics[0];
	*fy = intrinsics[4];
	*cx = intrinsics[2];
	*cy = intrinsics[5];
}

//---------------------------------------------------------

/**
	Convert raw depth value to meters.
	
	@param  raw_depth  raw depth value given by Kinect in [ 0 , 2047 ].
	@return  depth in meters.
*/
float rawDepthToMeters(int raw_depth)
{
	// 1053 <=> ~ 10 m
	// 1069 <=> ~ 20 m
	// 1074 <=> ~ 30 m
	// ...
	// 1084 <=> ~ 533 m
	// 1085 <=> ~ -836 m

	if (raw_depth < 1085)
	{
		return 1.0 / (raw_depth * -0.0030711016 + 3.3309495161);
	}
	
	return 0;
}

//---------------------------------------------------------

/**
	Unproject 2D point to 3D (reverse perspective transformation).
	
	@param  intrinsics  camera intrinsic parameters.
	@param  P2D  2D point coordinates.
	@param  distance  distance of 3D point.
	@return  3D point coordinates.
*/ 
CvPoint3D32f unproject( const float *intrinsics, const CvPoint *P2D, 
	float distance)
{
	CvPoint3D32f P3D;
	float fx, fy, cx, cy;

	getIntrinsicParameters( &fx, &fy, &cx, &cy, intrinsics);	

	P3D.x = (P2D->x - cx) * distance / fx;
	P3D.y = (P2D->y - cy) * distance / fy;
	P3D.z = distance;	// not exactly !

	return P3D;
}

//---------------------------------------------------------

/**
	Compute 3D point coordinates in new axis.
	
	@param  R  rotation of new axis in old axis.
	@param  T  translation of new axis in old axis.
	@param  P3D  3D point coordinates in old axis.
	@return  3D point coordinates in new axis.
*/ 
CvPoint3D32f changeAxis( const float *R, const float *T, CvPoint3D32f *P3D)
{
	CvPoint3D32f P3Dn; // 3D point in new axis
	
	// 3DPointInNewAxis = R.3DPointInOldAxis + T

	P3Dn.x = R[0]*P3D->x + R[1]*P3D->y + R[2]*P3D->z + T[0];
	P3Dn.y = R[3]*P3D->x + R[4]*P3D->y + R[5]*P3D->z + T[1];
	P3Dn.z = R[6]*P3D->x + R[7]*P3D->y + R[8]*P3D->z + T[2];
	
	return P3Dn;
}

//---------------------------------------------------------

/**
	Project 3D point to 2D (perspective transformation).
	
	@param  intrinsics  camera intrinsic parameters.
	@param  P3D  3D point coordinates.
	@return  2D point coordinates.
*/ 
CvPoint project( const float *intrinsics, CvPoint3D32f *P3D)
{
	CvPoint P2D;
	float fx, fy, cx, cy;

	getIntrinsicParameters( &fx, &fy, &cx, &cy, intrinsics);	
	
	P2D.x = (P3D->x * fx / P3D->z) + cx;
	P2D.y = (P3D->y * fy / P3D->z) + cy;
	
	return P2D;
}

//---------------------------------------------------------

/**
	Undistort camera image.
	
	@param  img  distorded image.
	@param  intrinsics  camera intrinsic parameters (3x3 matrix).
	@param  distortion  camera distortion coefficients (1x5 vector).
	@return  undistorded image.
*/ 
IplImage* undistort( IplImage *img, const float *intrinsics, 
	const float *distortion)
{
	CvMat intrinsicMat = cvMat( 3, 3, CV_32FC1, (void*) intrinsics);
	CvMat distortionCoeffs = cvMat( 1, 5, CV_32FC1, (void*) distortion);
	
	IplImage *undistortMapx = cvCreateImage( cvGetSize(img), IPL_DEPTH_32F,
		1 );
	IplImage *undistortMapy = cvCreateImage( cvGetSize(img), IPL_DEPTH_32F,
		1 );

	cvInitUndistortMap( &intrinsicMat, &distortionCoeffs, undistortMapx,
		undistortMapy);

	IplImage *undistordedImg = cvCloneImage(img);
	cvRemap( img, undistordedImg, undistortMapx, undistortMapy);
	
	cvReleaseImage(&undistortMapx);
	cvReleaseImage(&undistortMapy);
	
	return undistordedImg;
}

//---------------------------------------------------------

/**
	Compute color pixel matching depth pixel.
	
	@param  depth  depth image.
	@param  depth2DPoint  2D point coordinates in depth image.
	@return  2D point coordinates in color image, (0,0) if failed.
*/ 
CvPoint depthPointToColorPoint( IplImage *depth, CvPoint depth2DPoint)
{
	int rawDepth = 
		((int) cvGet2D( depth, depth2DPoint.y, depth2DPoint.x).val[0]) / 32;

	float distance = rawDepthToMeters(rawDepth);
	printf( "raw depth = %d, distance = %f m\n", rawDepth, distance);

	if( distance <= 0.1 )
		return cvPoint( 0, 0);
		// no valid depth data
	
	CvPoint3D32f depth3DPoint = unproject( depthIntrinsics, &depth2DPoint,
		distance);
	CvPoint3D32f color3DPoint = changeAxis( RDepthAxisToRgbAxis,
		TDepthAxisToRgbAxis, &depth3DPoint);
	CvPoint color2DPoint = project( rgbIntrinsics, &color3DPoint);
	
	return color2DPoint;
}


