#include <stdio.h>
#include <cv.h>
#include <highgui.h>

#include "kinectdepthconvert.h"
#include "depthToColorMapping.h"

//---------------------------------------------------------

struct KinectImages
{
	IplImage *depth; // raw depth image
	IplImage *cDepth; // color coded depth image
	IplImage *color; // color or gray image
};

//---------------------------------------------------------

void displayImageProperties(const IplImage *img)
{
	printf( "Image properties :\n");
	printf( " width    = %d\n", img->width);
	printf( " height   = %d\n", img->height);
	printf( " channels = %d\n", img->nChannels);
	printf( " depth    = ");
	if( img->depth == IPL_DEPTH_8U )
		printf("Unsigned 8-bit integer\n");
	else if( img->depth == IPL_DEPTH_8S )
		printf("Signed 8-bit integer\n");
	else if( img->depth == IPL_DEPTH_16U )
		printf("Unsigned 16-bit integer\n");
	else if( img->depth == IPL_DEPTH_16S )
		printf("Signed 16-bit integer\n");
	else if( img->depth == IPL_DEPTH_32S )
		printf("Signed 32-bit integer\n");
	else if( img->depth == IPL_DEPTH_32F )
		printf("Single-precision floating point\n");
	else if( img->depth == IPL_DEPTH_64F )
		printf("Double-precision floating point\n");
	printf( " datasize = %d bytes\n", img->imageSize);
}

//---------------------------------------------------------

void onMouse( int event, int x, int y, int flags, void* param)
{
	if( event != CV_EVENT_LBUTTONUP )
		return;
		
	struct KinectImages *imgs = (struct KinectImages *) param;

	CvPoint colorPoint = depthPointToColorPoint( imgs->depth, cvPoint( x, y));
	cvCircle( imgs->depth, cvPoint( x, y), 3, cvScalar(255));
	cvCircle( imgs->cDepth, cvPoint( x, y), 3, cvScalar(255));
	cvCircle( imgs->color, colorPoint, 3, cvScalar(255,0,0));
}

//---------------------------------------------------------

int main( int argc, const char **argv)
{
	if( argc != 4)
	{
		printf( "Syntaxe : %s  depth_image  color_image lut-file\n", argv[0]);

		return 1;
	}

	// load depth and color images	
		
	struct KinectImages kImgs;
	printf( "load depth image %s ...\n", argv[1]);
	kImgs.depth = cvLoadImage( argv[1], CV_LOAD_IMAGE_UNCHANGED);
	printf( "load color image %s ...\n", argv[2]);	
	kImgs.color = cvLoadImage( argv[2], CV_LOAD_IMAGE_UNCHANGED);

	printf( "Depth image properties :");
	displayImageProperties(kImgs.depth);
	printf( "Color image properties :");
	displayImageProperties(kImgs.color);
	
	// undistort images

	struct KinectImages uKImgs; // undistorded images
#if 0
	uKImgs.depth = undistort( kImgs.depth, depthIntrinsics, depthDistortion);
	uKImgs.color = undistort( kImgs.color, rgbIntrinsics, rgbDistortion);
#else
	// without undistortion
	uKImgs.depth = cvCloneImage(kImgs.depth);
	uKImgs.color = cvCloneImage(kImgs.color);
#endif

	// compute color coded depth images

	KinectDepthConvert *depthConverter = new KinectDepthConvert(argv[3]);
	kImgs.cDepth = depthConverter->convertToColor( kImgs.depth, false);
	uKImgs.cDepth = depthConverter->convertToColor( uKImgs.depth, false);
	if (uKImgs.cDepth==NULL)
	{
	    fprintf (stderr, "Converted depth image is NULL\n");
	    exit(1);
	}
	delete depthConverter;

	// init display

	cvNamedWindow( "depth", CV_WINDOW_AUTOSIZE );
	cvNamedWindow( "color", CV_WINDOW_AUTOSIZE );
	cvNamedWindow( "undistorded depth", CV_WINDOW_AUTOSIZE );
	cvNamedWindow( "color coded undistorded depth", CV_WINDOW_AUTOSIZE );
	cvNamedWindow( "undistorded color", CV_WINDOW_AUTOSIZE );


	// on click actions

	cvSetMouseCallback( "undistorded depth", onMouse, &uKImgs);
	cvSetMouseCallback( "color coded undistorded depth", onMouse, &uKImgs);

	// loop until user press 'q' or ESCAPE

	int key = 0;
	while( true )
	{
		// display images
		cvShowImage( "depth", kImgs.depth);
		cvShowImage( "color", kImgs.color);
		
		cvShowImage( "undistorded depth", uKImgs.depth);
		cvShowImage( "color coded undistorded depth", uKImgs.cDepth);
		cvShowImage( "undistorded color", uKImgs.color);
		
		key = 255 & cvWaitKey(5);
		if( key == 'q'  ||  key == 27 )
			break;
			
		if( key != 255 )
			printf( "key = %d\n", key);
	}
	
	// release memory
	
	cvDestroyAllWindows();
	cvReleaseImage( &(kImgs.depth));
	cvReleaseImage( &(kImgs.color));
	cvReleaseImage( &(kImgs.cDepth));
	cvReleaseImage( &(uKImgs.depth));
	cvReleaseImage( &(uKImgs.color));
	cvReleaseImage( &(uKImgs.cDepth));
}
